"""
## DACEyPy operators

This submodule defines some useful mathematical operators that can be imported
and applied directly to floats, NumPy arrays, DA scalars and DACEyPy arrays.
"""


import numpy as np

import daceypy as dp

_original_round = round


class ArgTypeException(TypeError):
    def __init__(self, x) -> None:
        if isinstance(x, (list, tuple)):
            msg = f"Invalid argument type: {type(x)} of {type(x[0])}"
        else:
            msg = f"Invalid argument type: {type(x)}"
        super().__init__(msg)


def DA(x):
    if isinstance(x, (dp.DA, float, int)):
        return dp.DA(x)
    raise ArgTypeException(x)


def array(x):
    if isinstance(x, (list, tuple, dp.array, np.ndarray)):
        return dp.array(x)
    raise ArgTypeException(x)


def round(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).round()
        if isinstance(x[0], (float, int)):
            return np.array(x).round()
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.round()
    if isinstance(x, (float, int)):
        return _original_round(x)
    raise ArgTypeException(x)


def root(x, p=2):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).root(p)
        if isinstance(x[0], (float, int)):
            return np.power(np.array(x), 1 / p)
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.root(p)
    if isinstance(x, (float, int)):
        return np.power(x, 1 / p)
    raise ArgTypeException(x)


def minv(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).minv()
        if isinstance(x[0], (float, int)):
            return 1 / np.array(x)
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.minv()
    if isinstance(x, (float, int)):
        return 1 / x
    raise ArgTypeException(x)


def sqr(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).sqr()
        if isinstance(x[0], (float, int)):
            return np.power(np.array(x), 2)
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.sqr()
    if isinstance(x, (float, int)):
        return np.power(x, 2)
    raise ArgTypeException(x)


def sqrt(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).sqrt()
        if isinstance(x[0], (float, int)):
            return np.sqrt(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.sqrt()
    if isinstance(x, (float, int)):
        return np.sqrt(x)
    raise ArgTypeException(x)


def isrt(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).isrt()
        if isinstance(x[0], (float, int)):
            return 1 / np.array(x)
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.isrt()
    if isinstance(x, (float, int)):
        return 1 / np.sqrt(x)
    raise ArgTypeException(x)


def cbrt(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).cbrt()
        if isinstance(x[0], (float, int)):
            return np.power(np.array(x), 1 / 3)
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.cbrt()
    if isinstance(x, (float, int)):
        return np.power(x, 1 / 3)
    raise ArgTypeException(x)


def icrt(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).icrt()
        if isinstance(x[0], (float, int)):
            return np.power(np.array(x), -1 / 3)
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.icrt()
    if isinstance(x, (float, int)):
        return np.power(x, -1 / 3)
    raise ArgTypeException(x)


def hypot(x, other):
    if isinstance(x, dp.DA):
        return x.hypot(other)
    if isinstance(x, (float, int)):
        return np.hypot(x, other)
    raise ArgTypeException(x)


def exp(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).exp()
        if isinstance(x[0], (float, int)):
            return np.exp(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.exp()
    if isinstance(x, (float, int)):
        return np.exp(x)
    raise ArgTypeException(x)


def log(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).log()
        if isinstance(x[0], (float, int)):
            return np.log(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.log()
    if isinstance(x, (float, int)):
        return np.log(x)
    raise ArgTypeException(x)


def logb(x, b=10.0):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).logb(b)
        if isinstance(x[0], (float, int)):
            return np.log(np.array(x)) / np.log(b)
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.logb(b)
    if isinstance(x, (float, int)):
        return np.log(x) / np.log(b)
    raise ArgTypeException(x)


def log10(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).log10()
        if isinstance(x[0], (float, int)):
            return np.log10(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.log10()
    if isinstance(x, (float, int)):
        return np.log10(x)
    raise ArgTypeException(x)


def log2(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).log2()
        if isinstance(x[0], (float, int)):
            return np.log2(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.log2()
    if isinstance(x, (float, int)):
        return np.log2(x)
    raise ArgTypeException(x)


def sin(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).sin()
        if isinstance(x[0], (float, int)):
            return np.sin(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.sin()
    if isinstance(x, (float, int)):
        return np.sin(x)
    raise ArgTypeException(x)


def cos(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).cos()
        if isinstance(x[0], (float, int)):
            return np.cos(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.cos()
    if isinstance(x, (float, int)):
        return np.cos(x)
    raise ArgTypeException(x)


def tan(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).tan()
        if isinstance(x[0], (float, int)):
            return np.tan(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.tan()
    if isinstance(x, (float, int)):
        return np.tan(x)
    raise ArgTypeException(x)


def asin(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).asin()
        if isinstance(x[0], (float, int)):
            return np.arcsin(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.asin()
    if isinstance(x, (float, int)):
        return np.arcsin(x)
    raise ArgTypeException(x)


def acos(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).acos()
        if isinstance(x[0], (float, int)):
            return np.arccos(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.acos()
    if isinstance(x, (float, int)):
        return np.arccos(x)
    raise ArgTypeException(x)


def atan(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).atan()
        if isinstance(x[0], (float, int)):
            return np.arctan(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.atan()
    if isinstance(x, (float, int)):
        return np.arctan(x)
    raise ArgTypeException(x)


def atan2(x, other):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).atan2(other)
        if isinstance(x[0], (float, int)):
            return np.arctan2(np.array(x), np.array(other))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.atan2(other)
    if isinstance(x, (float, int)):
        return np.arctan2(x, other)
    raise ArgTypeException(x)


def sinh(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).sinh()
        if isinstance(x[0], (float, int)):
            return np.sinh(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.sinh()
    if isinstance(x, (float, int)):
        return np.sinh(x)
    raise ArgTypeException(x)


def cosh(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).cosh()
        if isinstance(x[0], (float, int)):
            return np.cosh(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.cosh()
    if isinstance(x, (float, int)):
        return np.cosh(x)
    raise ArgTypeException(x)


def tanh(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).tanh()
        if isinstance(x[0], (float, int)):
            return np.tanh(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.tanh()
    if isinstance(x, (float, int)):
        return np.tanh(x)
    raise ArgTypeException(x)


def asinh(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).asinh()
        if isinstance(x[0], (float, int)):
            return np.arcsinh(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.asinh()
    if isinstance(x, (float, int)):
        return np.arcsinh(x)
    raise ArgTypeException(x)


def acosh(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).acosh()
        if isinstance(x[0], (float, int)):
            return np.arccosh(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.acosh()
    if isinstance(x, (float, int)):
        return np.arccosh(x)
    raise ArgTypeException(x)


def atanh(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).atanh()
        if isinstance(x[0], (float, int)):
            return np.arctanh(np.array(x))
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.atanh()
    if isinstance(x, (float, int)):
        return np.arctanh(x)
    raise ArgTypeException(x)


def erf(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).erf()
        if isinstance(x[0], (float, int)):
            return dp.array(x).erf().cons()
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.erf()
    if isinstance(x, (float, int)):
        return dp.DA(x).erf().cons()
    raise ArgTypeException(x)


def erfc(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).erfc()
        if isinstance(x[0], (float, int)):
            return dp.array(x).erfc().cons()
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.erfc()
    if isinstance(x, (float, int)):
        return dp.DA(x).erfc().cons()
    raise ArgTypeException(x)


def GammaFunction(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).GammaFunction()
        if isinstance(x[0], (float, int)):
            return dp.array(x).GammaFunction().cons()
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.GammaFunction()
    if isinstance(x, (float, int)):
        return dp.DA(x).GammaFunction().cons()
    raise ArgTypeException(x)


def LogGammaFunction(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).LogGammaFunction()
        if isinstance(x[0], (float, int)):
            return dp.array(x).LogGammaFunction().cons()
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.LogGammaFunction()
    if isinstance(x, (float, int)):
        return dp.DA(x).LogGammaFunction().cons()
    raise ArgTypeException(x)


def PsiFunction(x, n):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).PsiFunction(n)
        if isinstance(x[0], (float, int)):
            return dp.array(x).PsiFunction(n).cons()
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.PsiFunction(n)
    if isinstance(x, (float, int)):
        return dp.DA(x).PsiFunction(n).cons()
    raise ArgTypeException(x)


def cons(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], dp.DA):
            return dp.array(x).cons()
        raise ArgTypeException(x)
    if isinstance(x, (dp.DA, dp.array)):
        return x.cons()
    raise ArgTypeException(x)


def vnorm(x):
    if isinstance(x, dp.array):
        return x.vnorm()
    if isinstance(x, np.ndarray):
        return np.linalg.norm(x)
    raise ArgTypeException(x)



"""
# Differential Algebra Core Engine in Python - DACEyPy

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import annotations

from copy import copy
from typing import (Any, List, Mapping, Optional, Sequence, Tuple, Union,
                    overload)

import numpy as np
from numpy.typing import ArrayLike, NDArray

import daceypy

from ._PrettyType import PrettyType


class array(NDArray[np.object_], metaclass=PrettyType):
    """
    N-dimensional array of DA objects.
    """

    def __new__(cls, input_array) -> array:
        arr = np.asarray(input_array, dtype=np.object_)
        return number_to_DA_ufunc(arr).view(cls)

    def copy(self, order="K") -> array:
        """
        Create a copy of the DACEyPy array,
        creating a copy also of all the referenced DA objects.
        """
        return DA_copy_ufunc(super().copy(order))

    __copy__ = copy

    def __bytes__(self):
        raise NotImplementedError

    @staticmethod
    def fromText(
        text: str,
        alphabet: Union[
            Mapping[str, Union[float, daceypy.DA]],
            Sequence[str],
        ] = ("x", "y", "z", "xx", "yy", "zz"),
    ) -> array:
        """
        Compute a symbolic expression using DA.

        Args:
            text: string to evaluate.
            alphabet:
              sequence of strings that will be converted in order to DA vars,
              or dictionary with keys = names and vals = DA objects.

        Returns:
            Result of the evaluation as DA object.

        Raises:
            DACEException
            ValueError
        """

        out = daceypy.DA.fromText_general(text, alphabet)
        if isinstance(out, daceypy.DA):
            raise ValueError(
                "The expression returned a scalar instead of an array. "
                "Use daceypy.DA.fromText() instead of this method.")
        return out

    def vnorm(self) -> daceypy.DA:
        """
        Compute the Euclidean vector norm (length) of a DACEyPy array.

        Returns:
            A DA object containing the result of the operation.

        Derived from C++:
            `DA vnorm(const AlgebraicVector<DA> &obj)`
        """
        return self.sqr().sum().sqrt().item()

    def normalize(self) -> array:
        """
        Normalize the vector.

        Returns:
            A DACEyPy array of unit length.

        Raises:
            DACEException

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::normalize()`

        """
        return self / self.vnorm()

    def inv(self) -> array:
        """
        Compute the matrix inverse.
        Algorithm based on the Gauss elimination with full pivot
        (from the Numerical Cookbook).
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array that is the inverse of the original.

        Raises:
            ValueError

        Derived from C++:
            `AlgebraicMatrix<DA> AlgebraicMatrix<DA>::inv()`
        """
        # Check if matrix is 2D and square
        if self.ndim != 2 or self.shape[0] != self.shape[1]:
            raise ValueError("Matrix must be 2D square to compute the inverse")

        n = self.shape[0]
        k = 0
        det = daceypy.DA(1.0)
        R = list(range(n))
        P = [0] * n
        C1 = [0] * n
        C2 = [0] * n
        AA = self.copy()

        for i in range(n):
            k, err = _pivot(k, i, AA, P, R, C1, C2, det)
            if err:
                raise ValueError("Singolar matrix")
            _eliminate(k, AA, R)

        P = list(range(n))
        for i in range(n - 1, -1, -1):
            P[C1[i]], P[C2[i]] = P[C2[i]], P[C1[i]]

        AI = [[AA[R[i], P[j]] for j in range(n)] for i in range(n)]
        return array(AI)

    def det(self) -> daceypy.DA:
        """
        Compute the determinant of a square DACEyPy array.

        Returns:
            A DA object that is the determinant of the matrix.

        Raises:
            ValueError

        Derived from C++:
            `DA AlgebraicMatrix<DA>::det()`
        """
        # Check if matrix is 2D and square
        if self.ndim != 2 or self.shape[0] != self.shape[1]:
            raise ValueError("Matrix must be 2D square to compute the inverse")

        n = self.shape[0]
        k = 0
        det = daceypy.DA(1.0)
        R = list(range(n))
        P = [0] * n
        C1 = [0] * n
        C2 = [0] * n
        AA = self.copy()

        for i in range(n):
            k, err = _pivot(k, i, AA, P, R, C1, C2, det)
            if err:
                return daceypy.DA(0.0)
            _eliminate(k, AA, R)

        return det

    # *************************************************************************
    # *     Coefficient access routines
    # *************************************************************************

    def cons(self) -> NDArray[np.double]:
        """
        Extract the constant part of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array that contains the constant part of the original.

        Raises:
            DACEException

        Derived from C++:
            `AlgebraicMatrix<double> AlgebraicVector<DA>::cons()`
        """
        return DA_to_number_ufunc(self).view(np.ndarray)

    def linear(self) -> NDArray[np.double]:
        """
        Get the linear part of a DACEyPy array.

        Returns:
            A NumPy array with an additional dimension, that represents the DA
            variable of which the linear part is taken.

        Raises:
            DACEException

        Derived from C++:
            `AlgebraicMatrix<double> AlgebraicVector<DA>::linear()`
        """
        nvar = daceypy.DA.getMaxVariables()
        out = np.empty(self.shape + (nvar, ))
        for i in range(self.shape[0]):
            out[i] = self[i].linear()
        return out

    def concat(self, other: array) -> array:
        """
        Append a DACEyPy array to the end of the current one and return the new vector.

        Args:
            obj: the DACEyPy array to be appended.

        Returns:
            A new DACEyPy array containing the elements of both arrays.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::concat(const std::vector<DA> &obj)`
        """

        return np.concatenate((self, other)).view(array).copy()

    def trim(self, min: int, max: Optional[int] = None) -> array:
        """
        Returns a DACEyPy array with all monomials of order
        less than min and greater than max removed (trimmed).
        The result is copied in a new DACEyPy array.

        Args:
            min: minimum order to be preserved.
            max: maximum order to be preserved.

        Returns:
            A new DACEyPy array containing the result of the trimming.

        Raises:
            DACEException

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::trim(const unsigned int min, const unsigned int max = DA::getMaxOrder())`
        """
        if max is None:
            max = daceypy.DA.getMaxOrder()
        return np.frompyfunc(lambda da: daceypy.DA.trim(da, min, max), 1, 1)(self)

    # *************************************************************************
    # *     Math routines
    # *************************************************************************

    def cross(self, other) -> array:
        """
        Compute the cross product with another DACEyPy array.
        """
        return np.cross(self, other).view(array)

    def __pow__(self, p: Union[float, daceypy.DA]) -> array:  # type: ignore
        """
        Elevate a DACEyPy array to a given power.
        The result is copied in a new DACEyPy array.

        Args:
            p: power at which the DACEyPy array is elevated.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::pow(const int p)`
        """
        return np.power(self, p)  # type: ignore

    pow = __pow__

    def sqrt(self) -> array:
        """
        Compute the square root of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::sqrt()`
        """
        return np.sqrt(self)  # type: ignore[return-value]

    def exp(self) -> array:
        """
        Compute the exponent of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::exp()`
        """
        return np.exp(self)  # type: ignore[return-value]

    def log(self) -> array:
        """
        Compute the natural logarithm of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::log()`
        """
        return np.log(self)  # type: ignore[return-value]

    def sin(self) -> array:
        """
        Compute the sine of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::sin()`
        """
        return np.sin(self)  # type: ignore[return-value]

    def cos(self) -> array:
        """
        Compute the cosine of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::cos()`
        """
        return np.cos(self)  # type: ignore[return-value]

    def tan(self) -> array:
        """
        Compute the tangent of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::tan()`
        """
        return np.tan(self)  # type: ignore[return-value]

    def asin(self) -> array:
        """
        Compute the arcsine of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::asin()`
        """
        return np.arcsin(self)  # type: ignore[return-value]

    arcsin = asin

    def acos(self) -> array:
        """
        Compute the arccosine of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::acos()`
        """
        return np.arccos(self)  # type: ignore[return-value]

    arccos = acos

    def atan(self) -> array:
        """
        Compute the arctangent of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::atan()`
        """
        return np.arctan(self)  # type: ignore[return-value]

    arctan = atan

    def atan2(self, obj: array) -> array:
        """
        Compute the four-quadrant arctangent of Y/X. Y is the current vector,
        whereas X is the AlgebraicVector<DA> in input.
        The result is copied in a new DACEyPy array.

        Args:
            obj: AlgebraicVector<DA>

        Returns:
            A new DACEyPy array containing the result of the operation Y/X in [-pi, pi].

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::atan2(const AlgebraicVector<DA> &obj)`
        """
        return np.arctan2(self, obj)  # type: ignore[return-value]

    arctan2 = atan2

    def sinh(self) -> array:
        """
        Compute the hyperbolic sine of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::sinh()`
        """
        return np.sinh(self)  # type: ignore[return-value]

    def cosh(self) -> array:
        """
        Compute the hyperbolic cosine of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::cosh()`
        """
        return np.cosh(self)  # type: ignore[return-value]

    def tanh(self) -> array:
        """
        Compute the hyperbolic tangent of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::tanh()`
        """
        return np.tanh(self)  # type: ignore[return-value]

    def logb(self, b: float = 10.0) -> array:
        """
        Compute the logarithm of a DACEyPy array with respect to a given base.
        The result is copied in a new DACEyPy array.

        Args:
            b: base with respect to which the logarithm is computed (default = 10).

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::logb(const double b)`
        """
        return np.frompyfunc(lambda da: daceypy.DA.logb(da, b), 1, 1)(self)

    def log10(self) -> array:
        """
        Compute the base 10 logarithm of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.
        """
        return np.frompyfunc(lambda da: daceypy.DA.log10(da), 1, 1)(self)

    def log2(self) -> array:
        """
        Compute the base 2 logarithm of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.
        """
        return np.frompyfunc(lambda da: daceypy.DA.log2(da), 1, 1)(self)

    def isrt(self) -> array:
        """
        Compute the inverse square root of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::isrt()`
        """
        return np.frompyfunc(lambda da: daceypy.DA.isrt(da), 1, 1)(self)

    def cbrt(self) -> array:
        """
        Compute the cubic root of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.
        """
        return np.frompyfunc(lambda da: daceypy.DA.cbrt(da), 1, 1)(self)

    def icrt(self) -> array:
        """
        Compute the inverse cubic root of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.
        """
        return np.frompyfunc(lambda da: daceypy.DA.icrt(da), 1, 1)(self)

    def sqr(self) -> array:
        """
        Compute the square of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::sqr()`
        """
        return np.frompyfunc(lambda da: daceypy.DA.sqr(da), 1, 1)(self)

    def minv(self) -> array:
        """
        Compute the multiplicative inverse of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::minv()`
        """
        return np.frompyfunc(lambda da: daceypy.DA.minv(da), 1, 1)(self)

    def root(self, p: int) -> array:
        """
        Compute the p-th root of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Args:
            p: root to be computed (default = 2).

        Returns:
            A new DACEyPy array.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::root(const int p)`
        """
        return np.frompyfunc(lambda da: daceypy.DA.root(da, p), 1, 1)(self)

    def asinh(self) -> array:
        """
        Compute the hyperbolic arcsine of a DACEyPy array.
        The result is copied in a new DACEyPy array.


        Returns:
            A new DACEyPy array containing the result of the operation.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::asinh()`
        """
        return np.arcsinh(self)  # type: ignore[return-value]

    arcsinh = asinh

    def acosh(self) -> array:
        """
        Compute the hyperbolic arccosine of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array containing the result of the operation.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::acosh()`
        """
        return np.arccosh(self)  # type: ignore[return-value]

    arccosh = acosh

    def atanh(self) -> array:
        """
        Compute the hyperbolic arctangent of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array containing the result of the operation.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::atanh()`
        """
        return np.arctanh(self)  # type: ignore[return-value]

    arctanh = atanh

    def erf(self) -> array:
        """
        Compute the error function of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array containing the result of the operation.
        """
        return np.frompyfunc(lambda da: daceypy.DA.erf(da), 1, 1)(self)

    def erfc(self) -> array:
        """
        Compute the cumulative error function of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array containing the result of the operation.
        """
        return np.frompyfunc(lambda da: daceypy.DA.erfc(da), 1, 1)(self)

    def GammaFunction(self) -> array:
        """
        Compute the GammaFunction function of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array containing the result of the operation.
        """
        return np.frompyfunc(lambda da: daceypy.DA.GammaFunction(da), 1, 1)(self)

    def LogGammaFunction(self) -> array:
        """
        Compute the LogGammaFunction function of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Returns:
            A new DACEyPy array containing the result of the operation.
        """
        return np.frompyfunc(lambda da: daceypy.DA.LogGammaFunction(da), 1, 1)(self)

    def PsiFunction(self, n: int) -> array:
        """
        Compute the PsiFunction function of a DACEyPy array.
        The result is copied in a new DACEyPy array.

        Args:
            n: order of the Psi function (n >= 0).

        Returns:
            A new DACEyPy array containing the result of the operation.
        """
        return np.frompyfunc(lambda da: daceypy.DA.PsiFunction(da, n), 1, 1)(self)

    def deriv(self, p: int) -> array:
        """
        Compute the derivative of a DACEyPy array with respect to variable p.
        The result is copied in a new DACEyPy array.

        Args:
            p: variable with respect to which the derivative is calculated.

        Returns:
            A new DACEyPy array containing the result of the derivation.

        Raises:
            DACEException

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::deriv(const unsigned int p)`
        """
        return np.frompyfunc(lambda da: daceypy.DA.deriv(da, p), 1, 1)(self)

    def integ(self, p: int) -> array:
        """
        Compute the integral of a DACEyPy array with respect to variable p.
        The result is copied in a new DACEyPy array.

        Args:
            p: variable with respect to which the integral is calculated.

        Returns:
            A new DACEyPy array containing the result of the integration.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::integ(const unsigned int p)`
        """
        return np.frompyfunc(lambda da: daceypy.DA.integ(da, p), 1, 1)(self)

    def __setitem__(self, key, value) -> Any:
        """
        This is overridden to be sure to always use the already present
        DA objects in the array as assignment destination.
        """
        for dst_, val_ in np.nditer([self[key], value], ["refs_ok"]):
            # dst_ and val_ are np arrays of shape (1, )
            dst: daceypy.DA = dst_.item()  # get the DA object
            val = val_.item()  # get the object to be copied
            if isinstance(val, int):
                # if val is an int, we want to consider it a constant, not
                # the DA variable #val, so we cast it to a float
                val = float(val)
            # assign val to the content of the DA object dst
            dst.assign(val)

    # The following definitions are only to avoid type checking errors

    def __getitem__(self, key) -> Any:
        return super().__getitem__(key)

    def __add__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__add__(other).view(array)  # type: ignore

    def __sub__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__sub__(other).view(array)  # type: ignore

    def __mul__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__mul__(other).view(array)  # type: ignore

    def __matmul__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__matmul__(other).view(array)  # type: ignore

    def __truediv__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__truediv__(other).view(array)  # type: ignore

    def __radd__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__radd__(other).view(array)  # type: ignore

    def __rsub__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__rsub__(other).view(array)  # type: ignore

    def __rmul__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__rmul__(other).view(array)  # type: ignore

    def __rmatmul__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__rmatmul__(other).view(array)  # type: ignore

    def __rtruediv__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__rtruediv__(other).view(array)  # type: ignore

    def __iadd__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__iadd__(other).view(array)  # type: ignore

    def __isub__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__isub__(other).view(array)  # type: ignore

    def __imul__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__imul__(other).view(array)  # type: ignore

    def __itruediv__(self, other: Union[ArrayLike, daceypy.DA]) -> array:  # type: ignore
        return super().__itruediv__(other).view(array)  # type: ignore

    # *************************************************************************
    # *     Polynomial evaluation routines
    # *************************************************************************

    def compile(self) -> daceypy.compiledDA:
        """
        Compile vector of polynomials and create a compiledDA object.

        Returns:
            The compiled DA object.

        Derived from C++:
            `compiledDA AlgebraicVector<DA>::compile()`
        """
        return daceypy.compiledDA(self)

    @overload
    def eval(self, args: List[float]) -> List[float]:
        """
        Evaluate a DACEyPy array using a list of floats.

        Args:
            args: list of float to use for the evaluation.

        Returns:
            Result of the evaluation as list of floats.

        Raises:
            DACEException

        Derived from C++:
            `DA AlgebraicVector<DA>::eval(const std::vector<double> &args, std::vector<double> &res)`
        """
        ...

    @overload
    def eval(self, args: array) -> array:
        """
        Evaluate a DACEyPy array using a DACEyPy array.

        Args:
            args: DACEyPy array to use for the evaluation.

        Returns:
            Result of the evaluation as DACEyPy array.

        Raises:
            DACEException

        Derived from C++:
            `DA AlgebraicVector<DA>::eval(const std::vector<DA> &args, std::vector<DA> &res)`
        """
        ...

    @overload
    def eval(self, args: NDArray[np.double]) -> NDArray[np.double]:
        """
        Evaluate a DACEyPy array using a NumPy array of doubles.

        Args:
            args: NumPy array of doubles to use for the evaluation.

        Returns:
            Result of the evaluation as NumPy array of doubles.

        Raises:
            DACEException

        Derived from C++:
            `DA AlgebraicVector<DA>::eval(const std::vector<double> &args, std::vector<double> &res)`
        """
        ...

    @overload
    def eval(self, args: List[daceypy.DA]) -> List[daceypy.DA]:
        """
        Evaluate a DACEyPy array using a list of DA objects.

        Args:
            args: list of DA objects to use for the evaluation.

        Returns:
            Result of the evaluation as list of DA objects.

        Raises:
            DACEException

        Derived from C++:
            `DA AlgebraicVector<DA>::eval(const std::vector<DA> &args, std::vector<DA> &res)`
        """
        ...

    def eval(
        self,
        args: Union[
            List[float], List[daceypy.DA], NDArray[np.double], array],
    ) -> Union[List[float], List[daceypy.DA], NDArray[np.double], array]:
        return self.compile().eval(args)

    @overload
    def evalScalar(self, arg: float) -> NDArray[np.double]:
        """
        Evaluate a DACEyPy array with a single argument of type
        float and return vector of results.

        Args:
            arg:
              The value of the first independent DA variable to evaluate with.
              All remaining independent DA variables are assumed to be zero.

        Returns:
            NumPy array with the result of the evaluation.
        """
        ...

    @overload
    def evalScalar(self, arg: daceypy.DA) -> array:
        """
        Evaluate a DACEyPy array with a single argument of type
        DA and return vector of results.

        Args:
            arg: The value of the first independent DA variable to evaluate with.
              All remaining independent DA variables are assumed to be zero.

        Returns:
            DACEyPy array with the result of the evaluation.
        """
        ...

    def evalScalar(self, arg: Union[float, daceypy.DA]) \
            -> Union[NDArray[np.double], array]:
        args = np.array([arg]) if isinstance(arg, (float, int)) else array([arg])
        return self.eval(args)

    @overload
    def __call__(self, arg: float) -> NDArray[np.double]: ...

    @overload
    def __call__(self, arg: daceypy.DA) -> array: ...

    @overload
    def __call__(self, arg: NDArray[np.double]) -> NDArray[np.double]: ...

    @overload
    def __call__(self, arg: List[float]) -> List[float]: ...

    @overload
    def __call__(self, arg: array) -> array: ...

    @overload
    def __call__(self, arg: List[daceypy.DA]) -> List[daceypy.DA]: ...

    def __call__(
        self, arg: Union[
            float, daceypy.DA, NDArray[np.double],
            List[float], array, List[daceypy.DA],
        ]
    ) -> Union[NDArray[np.double], List[float], array, List[daceypy.DA]]:

        if isinstance(arg, (list, array, np.ndarray)):
            return self.eval(arg)
        return self.evalScalar(arg)

    def plug(self, var: int, val: float) -> array:
        """
        Partial evaluation of vector of polynomials. In each element of the vector,
        variable var is replaced by the value val. The resulting vector of DAs
        is returned.

        Args:
            var: variable number to be replaced.
            val: value by which to replace the variable.

        Returns:
            A new DACEyPy array containing the result of the operation.

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::plug(const unsigned int var, const double val)`
        """
        return np.frompyfunc(lambda da: daceypy.DA.plug(da, var, val), 1, 1)(self)

    def invert(self) -> array:
        """
        Invert the polynomials map given by the DACEyPy array.

        Returns:
            The inverted polynomials.

        Raises:
            TypeError
            Exception

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::invert()`
        """
        if self.ndim != 1:
            raise TypeError("This function works only on 1D objects")

        ord = daceypy.DA.getTO()
        nvar = len(self)

        if nvar > daceypy.DA.getMaxVariables():
            raise Exception(
                "DACE::AlgebraicVector<DA>::inverse: "
                "dimension of vector exceeds maximum number of DA variables.")

        # Create DA identity
        DDA = array.identity(nvar)

        # Split map into constant part AC,
        # non-constant part M, and non-linear part AN
        AC = self.cons()
        M = self.trim(1)
        AN = M.trim(2)

        # Extract the linear coefficients matrix
        AL = M.linear()

        # Compute the inverse of linear coefficients matrix
        AI = np.linalg.inv(AL)

        # Compute DA representation of the inverse of the linear part
        # of the map and its composition with non-linear part AN
        AIoAN_: array = AI @ AN  # type: ignore
        AIoAN = daceypy.compiledDA(AIoAN_)
        Linv = array(AI @ DDA)

        # Iterate to obtain the inverse map
        MI: array = Linv
        for i in range(ord):
            daceypy.DA.setTO(i + 1)
            MI = Linv - AIoAN.eval(MI)

        DDA_m_AC: array = DDA - AC
        return MI.eval(DDA_m_AC)

    # *************************************************************************
    # *     Input/Output routines
    # *************************************************************************

    def __str__(self) -> str:
        """
        Get a string representation of the DACEyPy array.

        Returns:
            Reference to output stream out.

        Derived from C++:
            `std::ostream& operator<<(std::ostream &out, const AlgebraicVector<DA> &obj)
            `std::ostream& operator<<(std::ostream &out, const AlgebraicMatrix<DA> &obj)
        """

        out: List[str] = []

        if self.ndim == 1:
            out.append(f"[[[ {len(self)} vector:\n")
            out.extend(f"{i}\n\n" for i in self)
            out.append("]]]")
        elif self.ndim == 2:
            nrows, ncols = self.shape
            out.append(f"[[[ {nrows}x{ncols} matrix:\n")
            for j in range(ncols):
                out.append(f"    Column {j + 1}\n")
                out.extend(f"{self[i, j]}\n" for i in range(nrows))
            out.append("]]]")
        else:
            return f"[[[ {'x'.join(map(str, self.shape))} DACEyPy array ]]]"

        return "".join(out)

    # *****************************************************************************
    # *     Auxiliary functions for ADS
    # *****************************************************************************

    def getTruncationErrors(self, type_: int = 0) -> NDArray[np.double]:
        """
        Return a NumPy array with truncation errors computed
        according to the given norm for each of the elements of the DA vector.

        Args:
            type_: type of the norm to be used, see documentation for DA.estimNorm.

        Raises:
            DACEException
        """
        errors = np.empty(self.shape[0])
        ord = daceypy.DA.getTO()
        el: daceypy.DA
        for i, el in enumerate(self):
            err, _ = el.estimNorm(0, type_, ord + 1)
            errors[i] = err[-1]
        return errors

    # *************************************************************************
    # *     Static factory routines
    # *************************************************************************

    @classmethod
    def identity(cls, n: Optional[int] = None) -> array:
        """
        Return the DA identity of dimension n.

        Args:
            n: dimension of the identity (default: n. of DA vars.).

        Raises:
            DACEException

        Derived from C++:
            `AlgebraicVector<DA> AlgebraicVector<DA>::identity(const size_t n)`
        """
        if n is None:
            n = daceypy.DA.getMaxVariables()
        return cls([daceypy.DA(i + 1) for i in range(n)])

    @classmethod
    def zeros(cls, shape: Union[int, Tuple[int, ...]]) -> array:
        """
        Return a DACEyPy array with null DA values.

        Args:
            shape: shape of the array.

        Raises:
            DACEException
        """
        return cls(np.zeros(shape))


# *****************************************************************************
# *     Auxiliary functions for inverse and determinant computation
# *****************************************************************************

def _pivot(
    k: int, ii: int, A: array, P: List[int], R: List[int],
    C1: List[int], C2: List[int], det: daceypy.DA,
) -> Tuple[int, bool]:
    """Auxiliary function for inverse and determinant computation"""

    im = t = m = 0
    n = A.shape[0]
    for i in range(n):
        if P[i] == 0:
            for j in range(n):
                if P[j] == 0:
                    t = abs(A[R[i], j])
                    if t >= m:
                        im = i
                        k = j
                        m = t
    if m < 1.e-12:
        return k, True
    else:
        det *= A[R[im], k]  # multiply pivot into determinant
        if im != k:
            det *= -1.0  # adjust sign of determinant
        R[im], R[k] = R[k], R[im]
        P[k] = 1  # mark column as done
        # record original row/column of pivot
        C1[ii] = im
        C2[ii] = k
    return k, False


def _eliminate(k: int, A: array, R: List[int]):
    """Auxiliary function for inverse and determinant computation"""

    n = A.shape[0]

    # # Better speed, less accuracy
    # for j in range(n):
    #     if j != k:
    #         A[R[k], j] = A[R[k], j] * A[R[k], k]

    # Better accuracy, less speed
    for j in range(n):
        if j != k:
            A[R[k], j] /= A[R[k], k]

    A[R[k], k] **= -1

    for i in range(n):
        if i != k:
            for j in range(n):
                if j != k:
                    A[R[i], j] -= A[R[i], k] * A[R[k], j]
            A[R[i], k] *= -A[R[k], k]


# *****************************************************************************
# *     Auxiliary functions for "vectorized" functions
# *****************************************************************************

def number_to_DA(n: float) -> daceypy.DA:
    return daceypy.DA.fromNumber(n)


def DA_to_number(da: daceypy.DA) -> float:
    return da.cons()


number_to_DA_ufunc = np.frompyfunc(number_to_DA, 1, 1)
DA_to_number_ufunc = np.vectorize(DA_to_number, [np.double])
DA_copy_ufunc = np.frompyfunc(copy, 1, 1)



"""
# Differential Algebra Core Engine in Python - DACEyPy

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import annotations

from typing import Tuple, Union, Optional, List

import numpy as np
from numpy.typing import NDArray

import daceypy

from .RK import RKCoeff, RK78

from ._integrator import integrator

from ._PrettyType import PrettyType

from ._ADS import ADS


class ADSstate(metaclass=PrettyType):
    """
    online ADS (Adaptive Domain Splitting) base element.
    """
    def __init__(
        self,
        ADSPatch: ADS,
        time: float = 0.0,
        stepsize: float = 0.0,
        checkBreached: bool = False,
        splitTimes: List[float] = [],
    ):
        """
        Initialize an instance of ADSstate class for propagation.

        Args:
            ADSPatch:
              instance of the class ADS that stores the current patch
              under analysis
            time: reference epoch for the patch
            stepsize: stepsize of the integrator at the patch epoch
            checkBreached: 
              check if patch needed split even though it was impossible
              during its propagation
            splitTimes: list of split times for the current patch

        See also:
            ADS
        """
        # ADSstate if a support class, its instances are used to store 
        # ADS info during propagation.

        self.ADSPatch: ADS = ADSPatch
        "The current ADS patch"
        self.time: float = time
        "Time at witch current patch started existing"
        self.stepsize: float = stepsize
        "Stepsize of the integrator when this patch was created"
        
        # moreover it has additional support variables to store ADS 
        # info in a easily accessible way.

        self.splitTimes: List[float] = splitTimes
        "List of split times for each patch"
        self.checkBreached: bool = checkBreached
        "Check if domain can be further split"
        # defined as "protected" because it should only be modified by 
        # a couple of class methods 
        self._breachTime: float = -1.0
        "Time at which split became necessary even though impossible"


class ADSintegrator(integrator, metaclass=PrettyType):
    """
    ADS integrator to perform splits online during propagation.
    """
    def __init__(self, RKcoeff: RKCoeff = RK78()) -> None:
        """
        Initialize an instance of the custom propagator capable of 
        dealing with online ADS.

        Args:
            RKcoeff:
              instance of the class RKCoeff that stores the coefficients
              of the Numerical Propagation Scheme

        See also:
            integrator
            RK.RKCoeff
        """

        # call to parent class constructor:
        # ADS is only available for daceypy.array type
        super(ADSintegrator, self).__init__(RKcoeff, daceypy.array)

        self._nSplitMax: int = 0
        "Maximum number of split for the ADS routine"

        self._errtol: Optional[Union[float, NDArray[np.double]]] = None
        "maximum truncation tolerance to determine split condition"
        
        self._stack: Optional[ADSstate] = None
        "ADS patch at the top of the stack with enhanced info"

    def _InitializeList(
        self,
        set: List[ADS],
        splitTimesList: Optional[List[List[float]]] = None,
    ) -> List[ADSstate]:
        """
        Initializes the list of ADS objects that need propagation as
        instances of ADSstate required by this custom propagator.

        Args:
            set: initial list of ADS objects.
            splitTimesList:
              optional list of split times of each element of set. 
              This is only required if one wants to propagate a domain
              that was already split while preserving its history.

        Returns:
            List of ADSstate instances ready for online ADS propagation

        Raises:
            ValueError
        """

        # NB: needs to be called after integrator.loadTime and 
        # integrator.loadStepSize
 
        out: List[ADSstate] = []

        # this part is only necessary if a pre-split domain is provided
        # and one wants to retain the split history (times).
        # If a list of split times is provided it must have proper size.
        # If it is not provided the info will be lost and only the new 
        # splits will be saved at the end of the propagation.
        if not splitTimesList:
            splitTimesList = [[]] * len(set)
        elif len(splitTimesList) != len(set):
            raise ValueError(
                "specified splitTimesList must have the same length"
                "as the number of elements in the initial set.")
        else:
            for i in range(len(set)):
                if len(set[i].nsplit) != len(splitTimesList[i]):
                    raise ValueError(
                        f"specified splitTimesList[{i}] must have the "
                        "same length as the number of splits in the "
                        "corresponding set element.")
    
        # create a list of ADSstate instances from the initial domains
        for i in range(len(set)):
            temp = ADSstate(
                set[i],
                self._t0,
                self._input.h,
                splitTimes = splitTimesList[i],
            )

            out.append(temp)

        # element at the top of the stack is stored in a variable
        self._stack = temp
    
        return out
    
    def _CallBack_CheckEvent(self) -> bool:
        """
        Check if integration should stop according to an event function
        (default is False). This version overloads that of integrator.

        Returns:
            True if the integration needs to be stopped

        See also:
            integrator._CallBack_CheckEvent
        """
        # function that overloads integrator.CallBack_CheckEvent:
        # If the current domain has already been tested and seen that 
        # cannot be split simply go ahead without doing anything
        if self._stack.checkBreached:
            self._checkStep = False
        else:
            if self._checkStep:
                # if the integration step was accepted check for split
                temp_f = ADS(
                    self._stack.ADSPatch.box,
                    self._stack.ADSPatch.nsplit,
                    self._runningX,
                )
                
                # if the domain can be split
                if temp_f.canSplit(self._nSplitMax): 
                    # If step is accepted and needs to be checked 
                    # but a violation is detected simply roll back 
                    # to previous time step
                    # If step is not accepted will not be checked
                    # and hence execution will never arrive here
                    if temp_f.checkSplit(self._errtol):
                        self._runningX = self._backX
                        self._input.t = self._backTime
                        self._input.h = self._backH
                    else:
                        self._checkStep = False
                else:
                    # the domain cannot be further split. 
                    # Set time of violation and violation flag 
                    # so that check can be avoided until tf 
                    self._stack.checkBreached = True
                    self._stack._breachTime = self._input.t
                    self._checkStep = False

        return self._checkStep
    
    def _SplitStateProcess(
        self,
        xf: daceypy.array,
        listIn: List[ADSstate],
        listOut: List[ADSstate],
    ) -> Tuple[List[ADSstate], List[ADSstate]]:
        """
        Updates the list of ADS objects that need propagation online.

        Args:
            xf: current state of the propagation.
            listIn: list of ADSstates that still need to be propagated.
            listOut: list of ADSstates that reached tf.

        Returns:
            List of ADSstate instances that still need propagation
            List of ADSstate instances that reached tf
            
        See also:
            ADS
        """
        # create temporary ads patch with latest propagator state
        state = ADS(
            self._stack.ADSPatch.box,
            self._stack.ADSPatch.nsplit,
            xf,
        )

        # if not at end of propagation this means that a split was 
        # necessary (and possible) from how we defined the checkevent
        if not self._reachstime:

            print(
                "The domain was split at instant ",
                self._input.t,
                " and continued the propagation...")
            
            # compute split subdomains
            dir = state.direction()
            Dl, Dr = state.split(dir)
            
            # update list of split times for each subdomain
            tempSplitTimes = self._stack.splitTimes.copy()
            tempSplitTimes.append(self._input.t)

            # add elements to the list of objects that need propagation
            tempL = ADSstate(
                Dl,
                self._input.t,
                self._input.h,
                splitTimes = tempSplitTimes,
            )
            tempR = ADSstate(
                Dr,
                self._input.t,
                self._input.h,
                splitTimes = tempSplitTimes,
            )
            
            # add one of the domains at the top of the stack
            self._stack = tempR
            # add other domain to list of element that need propagation
            listIn.append(tempL)
        else:
            # only here if reached tf (regardless of accuracy breach)
            tempD = ADSstate(
                state,
                self._input.t,
                self._input.h,
                self._stack.checkBreached,
                splitTimes = self._stack.splitTimes,
            )
            
            tempD._breachTime = self._stack._breachTime
            
            # add domain to final list: this step could be moved
            # outside of this function to avoid continuously passing 
            # listOut in and out of this function. It is only kept here
            # to simplify the propagate function but this could be
            # changed in the future for efficiency by simply passing
            # tempD as output.

            listOut.append(tempD)

        return listIn, listOut

    def _importfromList(self, ListIn: List[ADSstate]) -> None:
        """
        Remove from the list of ADS elements that need propagation the 
        one at the top of the stack and prepare it for propagation.

        Args:
            listIn: list of ADSstates that still need to be propagated.
            
        See also:
            integrator
        """
        # put first element to top of the stack
        self._stack = ListIn.pop()
        # pass its time and stepsize to the propagator
        self._input.t = self._stack.time
        self._input.h = self._stack.stepsize
        # set condition that it has not reached final time
        self._ReachsFinalTime() 

    def loadADSopt(
        self,
        tol: Union[float, NDArray[np.double]] = 1e-4,
        nsplit: int = 15,
    ) -> None:
        """
        Load ADS options.

        Args:
            tol: maximum truncation error tolerance.
            nsplit: maximum number of splits per each domain.
        """

        self._errtol = tol
        self._nSplitMax = nsplit

    def propagate(
        self,
        set: List[ADS],
        t1: float,
        t2: float,
        splitTimesList: List[List[float]] = [],
    ) -> List[ADSstate]:
        """
        Propagates the initial list of ADS objects while dealing with 
        online ADS. Overloads parent class method integrator.propagate.

        Args:
            List: list of ADS elements that need propagation.
            t1: initial reference time for each element of the List.
            t2: final time of the propagation.
            splitTimesList:
              optional list of split times of each element of set. 
              This is only required if one wants to propagate a domain
              that was already split while preserving its history.

        Returns:
            List of ADSstate instances that reached tf
            
        See also:
            integrator.propagate

        NB: t1 and t2 are not used in this function. They are only left
        here to maintain same syntax as the parent class method that 
        this function overloads. The initial time of the propagation 
        MUST be set through integrator.loadTime and it will be adapted
        online by each patch. Conversely, tf is fixed and it MUST be
        set through integrator.loadTime.
        """
        
        # list of final domains
        listOut: List[ADSstate] = []
        # create initial list from initial set of domains
        listIn = self._InitializeList(set, splitTimesList)

        print("Let's start propagation...")
        # loop until list of domains is not empty
        while listIn:
            # remove first domain and put it at the top of the stack
            self._importfromList(listIn)
            # loop until final time is reached for the domain
            while not self._reachstime:
                xf = super(ADSintegrator, self).propagate(self._stack.ADSPatch.manifold, self._input.t, t2)
                # after each call to propagation update list of patches
                listIn, listOut = self._SplitStateProcess(xf, listIn, listOut)

        print("Propagation completed. ")

        return listOut



"""
# Differential Algebra Core Engine in Python - DACEyPy

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict, Optional, overload

from . import core
from ._PrettyType import PrettyType


class DACEException(Exception, metaclass=PrettyType):

    @overload
    def __init__(self) -> None:
        """
        Create a DACEException object from an existing severity and error codes.

        Derived from C++:
            `DACEException::DACEException()`
        """
        ...

    @overload
    def __init__(self, exc_sv: int, exc_id: int) -> None:
        """
        Create a DACEException object from given severity and ID codes.

        Args:
            exc_sv: severity code of the error.
            exc_id: ID code of the error.

        Derived from C++:
            `DACEException::DACEException(const int exc_sv, const int exc_id)`
        """
        ...

    def __init__(
        self,
        exc_sv: Optional[int] = None,
        exc_id: Optional[int] = None,
    ) -> None:

        super().__init__()

        if exc_sv is not None:
            if exc_id is None:
                raise TypeError("exc_sv and exc_id must be defined together")
            if not isinstance(exc_sv, int):
                raise TypeError("exc_sv must be an int")
            if not isinstance(exc_id, int):
                raise TypeError("exc_id must be an int")
            self.m_x = exc_sv
            self.m_yy = exc_id
        elif exc_id is not None:
            raise TypeError("exc_sv and exc_id must be defined together")
        else:
            self.m_x = core.GetErrorX()
            self.m_yy = core.GetErrorYY()

        self.updateMessage()

    def __str__(self):
        return self.msg

    # *************************************************************************
    # *     Private member functions
    # *************************************************************************

    def updateMessage(self):
        """
        Update the error message of this exception based on its ID.
        """

        self.id = self.m_x * 100 + self.m_yy

        if self.m_x > 10:
            self.msg = f"{DACEerr[self.id]} (ID: {self.id})"
        else:
            self.msg = f"{core.GetErrorMSG().decode()} (ID: {self.id})"
            core.ClearError()


DACEerr: Dict[int, str] = {
    0: "DACE: Unknown DACE error. Contact Dinamica SRL for filing a bug report.",

    # 101: "DACEDAL: Attempt to deallocate protected or unallocated variable, ignored",
    # 102: "DACESETNOT: Truncation order set to value larger then maximum order, setting to maximum order",
    # 103: "DACEEST: Not enough non-zero monomials found, returned estimate may be inaccurate",
    # 104: "DACEREAD: Line numbering out of order, ignoring line numbers",
    # 105: "DACEREAD: DA vector contains more variables than current setup, ignoring coefficient",
    # 106: "DACEREAD: Duplicate monomial in DA vector",
    # 107: "DACEINI: Requested order has been increased to the minimum required order of 1",
    # 108: "DACEINI: Requested number of variables has been increased to the minimum required number of variables 1",

    # 601: "DACEVAR: Requested independent variable is out of bounds of current setup, returning zero DA",
    # 602: "DACEPOK: Not enough storage to insert monomials, truncating",
    # 603: "DACEVAR: Not enough storage to set up variable, returning zero DA",
    # 604: "DACECOEF: Not enough storage to set coefficient, truncating",
    # 605: "DACEDIVC: Divide by zero, returning zero DA",
    # 606: "DACEINT: Requested independent variable out of bounds of current setup, returning zero DA",
    # 607: "DACEDER: Requested independent variable out of bounds of current setup, returning zero DA",
    # 608: "DACEMINV: Divide by zero, returning zero DA",
    # 609: "DACESQRT: Negative constant part in square root, returning zero DA",
    # 610: "DACEISQRT: Negative constant part in inverse square root, returning zero DA",
    # 611: "DACELOG: Negative constant part in logarithm, returning zero DA",
    # 612: "DACETAN: Cosine is zero in tangent, returning zero DA",
    # 613: "DACEASIN: Constant part is out of domain [-1,1] in arcsine, returning zero DA",
    # 614: "DACEACOS: Constant part is out of domain [-1,1] in arccosine, returning zero DA",
    # 617: "DACEACOSH: Constant part is out of domain [1,infinity) in hyperbolic arccosine, returning zero DA",
    # 618: "DASEATANH: Constant part is out of domain [-1,1], returning zero DA",
    # 619: "DACEONORM: Requested independent variable out of bounds of current setup, returning zero",
    # 620: "DACEEST: Maximum order must be at least 2 in order to use DA estimation",
    # 622: "DACEPLUG: Requested independent variable out of bounds of current setup, returning zero DA",
    # 623: "DACELOGB: Logarithm base must be positive, returning zero DA",
    # 624: "DACEREAD: Not enough lines provided to read a DA vector, returning zero DA",
    # 625: "DACEREAD: Unrecognized DA input format, returning zero DA",
    # 627: "DACEENC: Invalid exponents with order larger than maximum order, returning zero",
    # 628: "DACEDEC: Invalid DA codes provided, returning zero",
    # 629: "DACEROOT: Zero-th root does not exists, returning zero DA",
    # 630: "DACEROOT: Negative or zero constant part in even root, returning zero DA",
    # 631: "DACEROOT: Zero constant part in odd root, returning zero DA",
    # 632: "DACEPAC: Not enough storage in the target object, truncating",
    # 633: "DACECMUL: Not enough storage in the target object, truncating",
    # 634: "DACELIN: Not enough storage in the target object, truncating",
    # 635: "DACEINT: Not enough storage in the target object, truncating",
    # 636: "DACEDER: Not enough storage in the target object, truncating",
    # 637: "DACECOP: Not enough storage in the target object, truncating",
    # 638: "DACEPUSH: Not enough space to store the provided data, truncating",
    # 639: "DACEPULL: Not enough space to store the requested data, truncating",
    # 640: "DACETRIM: Not enough space to store the requested data, truncating",
    # 641: "DACENORM: Unknown norm type, resorting to max norm",
    # 642: "DACEONORM: Unknown norm type, resorting to max norm",
    # 643: "DACETRIM: Not enough storage in the target object, truncating",
    # 644: "DACEWRITE: DA vector contains more monomials than expected, truncating",

    # 901: "DACEINI: Internal array size exceeds available addressing space",

    # 98: "DACEINI: Internal error, size of generated DA coding arrays does not match number of monomials",
    # 97: "DACEINI: Internal error, generated DA coding arrays are faulty",
    # 96: "DACEINI: Internal error, unable to correctly allocate scratch variables",
    # 94: "DACEINI: Internal error, memory allocation failure",
    # 93: "DACEREALL: Internal error, memory allocation failure",
    # 92: "DACEALL: DACE has not been initialized",
    # 90: "DACEINFO: DA Object not allocated",

    101: "DA::getCoeff: Not enough exponents, missing exponents treated as zero",
    102: "DA::setCoeff: Not enough exponents, missing exponents treated as zero",
    103: "DA::getCoeff: More exponents than variables, ignoring extra exponents",
    104: "DA::setCoeff: More exponents than variables, ignoring extra exponents",

    604: "compiledDA::compiledDA(): Dimension lower than 1",
    605: "compiledDA::eval: Argument size lower than the number of variables in the polynomial",

    99: "DA::checkVersion: DACE C++ interface header file and DACE core library version mismatch",
}



"""
# Differential Algebra Core Engine in Python - DACEyPy

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import annotations

from typing import Callable, Optional, Tuple, Type, Union, overload

import numpy as np
from numpy.typing import NDArray

import daceypy

from ._PrettyType import PrettyType
from .RK import RK78, RKCoeff, RKstate


class integrator(metaclass=PrettyType):
    """
    Modular integrator class for DA objects
    """

    # *************************************************************************
    # *     Constructors & Destructors
    # *************************************************************************

    def __init__(
        self,
        RKcoeff: Optional[RKCoeff] = RK78(),
        stateType: Optional[Type] = np.ndarray,
    ) -> None:
        """
        Initialize a propagator object.

        Args:
            RKcoeff:
                instance of the class RKCoeff that stores the coefficients
                of the Numerical Propagation Scheme

        See also:
            RK.RKCoeff
            DA.getMonomial
        """

        # declaration of protected instance variables
        self._runningX: Union[daceypy.array, NDArray[np.double]] = None
        "current state"
        self._backX: Union[daceypy.array, NDArray[np.double]] = None
        "backward state"
        self._outX: Union[daceypy.array, NDArray[np.double]] = None
        "final state"
        self._stateType: Type = stateType

        self._t0: float = None
        "initial time"
        self._tf: float = None
        "final time (if I build well the Integrator it could be also befor of t0)"
        self._DT: float = None
        "support variable to store tf-t0 and avoid multiple type checks "
        self._propDir: int = 0
        "propagation direction (forward =1, backward=-1)"

        self._backTime: float = None
        "backward time"
        self._runningTime: float = None
        "current time"
        self._tout: float = None
        "final time of propagation"

        self._backH: float = None
        "backward timestep "
        self._epstol: NDArray[np.double] = None
        "Step tolerance"
        self._err: float = None
        "Error step tolerance"

        self._checkStep: bool = None
        "check if propagation step is accepted or if stepsize needs adjusting"
        self._reachstime: bool = None
        "check if final time is reached"

        self._input: RKstate = RKstate()
        "support object to store current state of the propagator"

        self._RKcoeff: RKCoeff = RKcoeff
        "support object to store Butcher's tableau of the propagation method"

        # perform dynamic assignment to avoid continous type checks
        if stateType is np.ndarray:
            setattr(integrator, '_geterr', _geterrFloat)
            setattr(integrator, '_getepstol', _getepstolFloat)
            self._inputType = np
        elif stateType is daceypy.array:
            setattr(integrator, '_geterr', _geterrDA)
            setattr(integrator, '_getepstol', _getepstolDA)
            self._inputType = daceypy.array

    # *************************************************************************
    # *     User Overloadable Routines
    # *************************************************************************

    def _CallBack_getStepSize(self) -> float:
        """
        Provides a user defined scaling of the optimal integration step
        (default = 1). Can be overloaded by the user.

        Returns:
            a user defined scaling factor (default is 1)
        """
        return self._DT

    def _CallBack_CheckEvent(self) -> bool:
        """
        Check if integration should stop according to an event function
        (default is False).
        Can be overloaded by the user.

        Returns:
            True if the integration needs to be stopped
        """
        if (self._checkStep):
        #User could define some event ot check within their own derived class
            self._checkStep = False
        return False

    def _acceptStep(
        self,
        pn: Union[daceypy.array, NDArray[np.double]],
        tnew: float,
    ) -> None:
        """
        Updates the integration state if the tolerances are satisfied.
        Can be overloaded by the user to define a custom set of actions.

        Args:
            pn: propagated state (either numpy.ndarray or daceypy.array)
            tnew: integration time reached
        """
        if self._err <= 1.0:
            self._checkStep = True
            self._runningX = pn
            self._input.t = tnew
            self._runningTime = tnew

    @staticmethod
    def f(
        x: Union[daceypy.array, NDArray[np.double]],
        t: float,
    ) -> Union[daceypy.array, NDArray[np.double]]:
        """
        Computes the righ-hand-side of the ODE.
        Must be overloaded by the user.

        Args:
            x: state (either numpy.ndarray or daceypy.array)
            t: time (float)

        Returns:
            The time derivatives of the state x at time t
            (either as numpy.ndarray or daceypy.array)
        """
        raise NotImplementedError()

    # *************************************************************************
    # *     Support Routines
    # *************************************************************************

    def _ReachsFinalTime(self) -> bool:
        """
        Checks if integration has reached final time.

        Returns:
            True if the integration should stop.
        """
        self._reachstime = (abs(1.0 - self._input.t / self._tf) <= 2.2e-16)
        return self._reachstime

    def _Initialize(
        self,
        Initset: Union[daceypy.array, NDArray[np.double]],
        t1: float,
        t2: float,
    ) -> None:
        """
        Initializes the state and time of the propagator.

        Args:
            Initset: initial state (either numpy.ndarray or daceypy.array)
            t1: initial time of the propagation
            t2: final time of the propagation

        Raises:
            TypeError
        """
        self._runningX = Initset
        self._runningTime = t1
        self._input.t = t1
        if type(self._runningX) is not self._stateType:
            raise TypeError(
                "propagator was initialized to propagate "
                f"(\"{self._stateType}\") however, "
                f"(\"{type(self._runningX)}\") was given")

    def _computeStep(self) -> \
            Tuple[Union[daceypy.array, NDArray[np.double]], Union[daceypy.array, NDArray[np.double]]]:
        """
        Computes the state after one propagation step according to the
        selected numerical scheme.

        Returns:
            Two states propagated at different orders to check error for stepsize adaptation.
        """
        Y = self._inputType.zeros((self._runningX.size, self._RKcoeff.RK_stage))
        Y[:, 0] = self._runningX.copy()
        feval = self.f(Y[:, 0], self._input.t + self._input.h * self._RKcoeff.gamma[0])
        pn = self._runningX + self._input.h * self._RKcoeff.beta[0] * feval
        pndiff = (self._RKcoeff.beta_star[0] - self._RKcoeff.beta[0]) * feval

        ia = 0
        for i in range(1, self._RKcoeff.RK_stage):
            Yk = self._inputType.zeros(self._runningX.size)
            for j in range(i):
                tj = self._input.t + self._input.h * self._RKcoeff.gamma[j]
                Yk += self._RKcoeff.alpha[ia] * self.f(Y[:, j], tj)
                ia += 1

            Y[:, i] = self._runningX + self._input.h*Yk
            feval = self.f(Y[:, i], self._input.t + self._input.h*self._RKcoeff.gamma[i])
            pn += self._input.h*self._RKcoeff.beta[i] * feval
            pndiff += (self._RKcoeff.beta_star[i] - self._RKcoeff.beta[i]) * feval
        pndiff *= self._input.h

        return pn, pndiff

    @overload
    def _geterr(self, pndiff: Union[daceypy.array, NDArray[np.double]]) -> float:
        """
        Computes the propagation error for the given timestep.

        Args:
            pndiff:
                Variation of state caused by different propagation orders.

        Returns:
            The normalized error to determine stepsize adaptation
        """
        ...

    @overload
    def _getepstol(self, pn: Union[daceypy.array, NDArray[np.double]]) -> NDArray[np.double]:
        """
        Computes the error tolerance for each step.

        Args:
            pn: state propagated according to current timestep.

        Returns:
            A vector of tolerances for each element of the state.
        """
        ...

    def _adaptStepSize(self) -> None:
        """
        Adapt the stepsize according to step error.
        """

        if self._err == 0.0:
            self._input.h=4.0
        else:
            self._input.h *= min(4.0, max(0.1, 0.9 * pow(1.0 / self._err, 1.0 / (self._RKcoeff.RK_order + 1.0))))
        

        if abs(self._input.h) <= self._input.minh * 1.2:
            self._input.h /= 3.0
            print(" --- WARNING MINIMUM STEP SIZE REACHED.")

        if abs(self._input.h) > self._input.maxh:
            self._input.h = self._propDir*self._input.maxh

        dt = abs(self._tf - self._input.t)
        if dt < abs(self._input.h):
            self._input.h = self._propDir*dt

    # *************************************************************************
    # *     Functions that should be accessed outside class
    # *************************************************************************
    def loadTime(self, t1: float, t2:  float) -> None:
        """
        Loads initial and final propagation times inside an instance of integrator class.

        Args:
            t1: initial time
            t2: final time
        """
        if not isinstance(t1, float):
            if isinstance(t1, int):
                t1=float(t1)
            else:
                raise TypeError(f"t1 must be a float (\"{type(t1)}\") was given")
        if not isinstance(t2, float):
            if isinstance(t2,int):
                t2=float(t2)
            else:
                raise TypeError(f"t2 must be a float (\"{type(t2)}\") was given")

        self._t0 = t1
        self._tf = t2

        self._DT = self._tf - self._t0
        self._propDir = 1 if self._DT > 0 else -1

    def loadTol(self, tol1: float = 0.0, tol2: float = 0.0) -> None:
        """
        Loads absolute and relative tolerances of integrator.

        Args:
            tol1: absolute tolerance
            tol2: relative tolerance
        """
        self._input.epsabs = tol1
        self._input.epsrel = tol2 or 1e-10

    def loadStepSize(self, h0: float = 0.0, max: float = 0.0, min: float = 0.0) -> None:
        """
        Loads initial, maximum, and minimum stepsize of integrator.

        Args:
            h0: initial stepsize
            max: maximum stepsize
            min: minimum stepsize
        """
        self._input.h = h0
        self._input.maxh = max
        self._input.minh = min

        if self._input.h == 0.0:
            self._input.h = self._propDir * np.log10(abs(self._DT) * 0.5 + 1.0)
        else:
            self._input.h = self._propDir * abs(self._input.h)

        self._input.h0 = self._input.h

        if self._input.maxh == 0.0:
            self._input.maxh = 0.4 * abs(self._DT)
        if self._input.minh == 0.0:
            self._input.minh = self._input.maxh * 2.2e-16

    def propagate(self, Initset: Union[daceypy.array, NDArray[np.double]], t1: float, t2: float) -> Union[daceypy.array, NDArray[np.double]]:
        """
        Propagates state in time interval according to selected numerical scheme.

        Args:
            Initset: initial state
            t1: initial time
            t2: final time

        Returns:
            The final state of the integration
        """
        self._Initialize(Initset, t1, t2)
        self._checkStep = False

        while not self._ReachsFinalTime() and not self._CallBack_CheckEvent():
            # ask user if wants to use dt as step size or not, as default dt is equal to optimal h
            dt = self._CallBack_getStepSize()
            self._input.h = abs(self._input.h) * dt / abs(self._DT)

            self._backX = self._runningX
            self._backTime = self._input.t
            self._backH = self._input.h
            v1, v1diff = self._computeStep()
            self._epstol = self._getepstol(v1)
            self._err = self._geterr(v1diff)
            self._acceptStep(v1, self._input.t + self._input.h)
            self._adaptStepSize()

        self._tout = self._input.t
        self._outX = self._runningX
        return self._runningX


# *************************************************************************
# *     Functions specialized for float type
# *************************************************************************

def _geterrFloat(self, pndiff: NDArray[np.double]) -> float:
    """
    Computes the propagation error for the given timestep.

    Args:
        pndiff:
            Variation of state caused by different propagation orders.

    Returns:
        The normalized error to determine stepsize adaptation
    """

    E: float = 0.0
    size: int = self._runningX.size

    temp = pow(pndiff / self._epstol, 2.0)
    E = sum(temp)

    return np.sqrt(E/size)

def _getepstolFloat(self, pn: NDArray[np.double]) -> NDArray[np.double]:
    """
    Computes the error tolerance for each step.

    Args:
        pn: state propagated according to current timestep.

    Returns:
        A vector of tolerances for each element of the state.
    """

    vconst = self._runningX
    v1const = pn

    tol = np.maximum(abs(vconst), abs(v1const))

    tol = self._input.epsabs + tol*self._input.epsrel

    return tol

# *************************************************************************
# *     Functions specialized for DA type
# *************************************************************************

def _geterrDA(self, pndiff: daceypy.array) -> float:
    """
    Computes the propagation error for the given timestep.

    Args:
        pndiff:
            Variation of state caused by different propagation orders.

    Returns:
        The normalized error to determine stepsize adaptation
    """

    E: float = 0.0
    size: int = self._runningX.size

    temp = pndiff.cons()

    E = sum(pow(temp/self._epstol, 2.0))

    return np.sqrt(E/size)


def _getepstolDA(self, pn: daceypy.array) -> NDArray[np.double]:
    """
    Computes the error tolerance for each step.

    Args:
        pn: state propagated according to current timestep.

    Returns:
        A vector of tolerances for each element of the state.
    """

    vconst = self._runningX.cons()
    v1const = pn.cons()

    tol = np.maximum(abs(vconst), abs(v1const))

    tol = self._input.epsabs + tol*self._input.epsrel

    return tol


def PicardLindelof(
    x: Union[daceypy.array, NDArray[np.double]],
    direction: int, 
    tf: float,
    f: Callable[[daceypy.array, Union[daceypy.DA, float]], daceypy.array],
) -> daceypy.array:
    """
    Computes the time expansion at final time with Picard Lindelof operator

    Args:
        x: expansion state.
        direction: index of DA variable corresponding to time.
        tf: constant float expressing the reference time of expansion.
        f: right hand side of ODE.

    Returns:
        The input state expanded in time about input reference time.
    """
    expansionOrder = daceypy.DA.getMaxOrder()
    t = tf + daceypy.DA(direction)

    x_i = x.copy()

    for i in range(expansionOrder):
        # iteration n times
        x_i = x + f(x_i, t).integ(direction)

    return x_i



"""
# Differential Algebra Core Engine in Python - DACEyPy

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from ctypes import (POINTER, Structure, c_bool, c_char_p, c_double, c_int,
                    c_uint, c_void_p, cdll)
from inspect import getsourcefile
from pathlib import Path
from typing import Callable

from . import _DACEException
from .get_platform import get_platform

sourcefile = getsourcefile(lambda: 0)
assert sourcefile is not None
libfolder = Path(sourcefile).resolve().parent / "lib"
system, machine = get_platform()

try:
    library = next(libfolder.glob(f"dace_{system}-{machine}*"))
except StopIteration:
    raise Exception(
        f"DACEyPy does not support this architecture ({system} {machine}).\n"
        "Supported architectures:\n" + "\n".join(
            " - " + lib.stem[5:].replace("-", " ")
            for lib in libfolder.glob(f"dace_*")))

DAlib = cdll.LoadLibrary(str(library))


class dmonomial(Structure):
    __slots__ = "cc", "ii"
    __fields__ = [("cc", c_double), ("ii", c_uint)]


class DACEDA(Structure):

    __slots__ = "len", "max", "dmonomial"

    _fields_ = [
        ("len", c_uint),
        ("max", c_uint),
        ("dmonomial", POINTER(dmonomial)),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        AllocateDA(self, 0)

    def __del__(self):
        FreeDA(self)


c_int_p = POINTER(c_int)
c_uint_p = POINTER(c_uint)
c_double_p = POINTER(c_double)
DACEDA_p = POINTER(DACEDA)
DACEDA_p_p = POINTER(DACEDA_p)


def corefunction(argtypes=[], restype=None, errcheck: bool = True):
    """
    Provides a function decorator that is used to get the corresponding
    function from the DA core dynamic-link library.
    """

    def decorator(fun: Callable) -> Callable:
        function = getattr(DAlib, "dace" + fun.__name__)
        function.restype = restype
        function.argtypes = argtypes

        if not errcheck:
            # just return the function
            return function

        # wrap the function to automatically check for errors
        def with_errcheck(*args):
            res = function(*args)
            if GetError():
                raise _DACEException.DACEException
            return res

        return with_errcheck

    return decorator


# *****************************************************************************
# *     DACE initialization and state related routines
# *****************************************************************************


@corefunction([c_uint, c_uint])
def Initialize(no: int, nv: int) -> None:
    """`void daceInitialize(unsigned int no, unsigned int nv);`"""
    ...


@corefunction()
def InitializeThread() -> None:
    """`void daceInitializeThread();`"""
    ...


@corefunction()
def CleanupThread() -> None:
    """`void daceCleanupThread();`"""
    ...


@corefunction([c_int_p, c_int_p, c_int_p])
def GetVersion(imaj, imin, ipat) -> None:
    """`void daceGetVersion(int REF(imaj), int REF(imin), int REF(ipat));`"""
    ...


@corefunction([c_double], c_double)
def SetEpsilon(deps: float) -> float:
    """`double daceSetEpsilon(const double deps);`"""
    ...


@corefunction([], c_double)
def GetEpsilon() -> float:
    """`double daceGetEpsilon();`"""
    ...


@corefunction([], c_double)
def GetMachineEpsilon() -> float:
    """`double daceGetMachineEpsilon();`"""
    ...


@corefunction([], c_uint)
def GetMaxOrder() -> int:
    """`unsigned int daceGetMaxOrder();`"""
    ...


@corefunction([], c_uint)
def GetMaxVariables() -> int:
    """`unsigned int daceGetMaxVariables();`"""
    ...


@corefunction([], c_uint)
def GetMaxMonomials() -> int:
    """`unsigned int daceGetMaxMonomials();`"""
    ...


@corefunction([], c_uint)
def GetTruncationOrder() -> int:
    """`unsigned int daceGetTruncationOrder();`"""
    ...


@corefunction([c_uint], c_uint)
def SetTruncationOrder(fnot: int) -> int:
    """`unsigned int daceSetTruncationOrder(const unsigned int fnot);`"""
    ...


# *****************************************************************************
# *     DACE error state routine
# *****************************************************************************


@corefunction([], c_uint, errcheck=False)
def GetError() -> int:
    """`unsigned int daceGetError();`"""
    ...


@corefunction([], c_uint, errcheck=False)
def GetErrorX() -> int:
    """`unsigned int daceGetErrorX();`"""
    ...


@corefunction([], c_uint, errcheck=False)
def GetErrorYY() -> int:
    """`unsigned int daceGetErrorYY();`"""
    ...


@corefunction([], c_char_p, errcheck=False)
def GetErrorFunName() -> bytes:
    """`const char* daceGetErrorFunName();`"""
    ...


@corefunction([], c_char_p, errcheck=False)
def GetErrorMSG() -> bytes:
    """`const char* daceGetErrorMSG();`"""
    ...


@corefunction(errcheck=False)
def ClearError() -> None:
    """`void daceClearError();`"""
    ...


# *****************************************************************************
# *     DACE memory handling routines
# *****************************************************************************


@corefunction([DACEDA_p, c_uint])
def AllocateDA(inc, len: int) -> None:
    """`void daceAllocateDA(DACEDA REF(inc), const unsigned int len);`"""
    ...


@corefunction([DACEDA_p])
def FreeDA(inc) -> None:
    """`void daceFreeDA(DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p])
def InvalidateDA(inc) -> None:
    """`void daceInvalidateDA(DACEDA REF(inc));`"""
    ...


@corefunction()
def MemoryDump() -> None:
    """`void daceMemoryDump();`"""
    ...


# *****************************************************************************
# *     DACE variable creation routines
# *****************************************************************************


@corefunction([DACEDA_p, c_uint, c_double])
def CreateVariable(ina, i: int, ckon: float) -> None:
    """`void daceCreateVariable(DACEDA REF(ina), const unsigned int i, const double ckon);`"""
    ...


@corefunction([DACEDA_p, c_uint_p, c_double])
def CreateMonomial(ina, jj, ckon: float) -> None:
    """`void daceCreateMonomial(DACEDA REF(ina), const unsigned int jj[], const double ckon);`"""
    ...


@corefunction([DACEDA_p, c_double])
def CreateConstant(ina, ckon: float) -> None:
    """`void daceCreateConstant(DACEDA REF(ina), const double ckon);`"""
    ...


@corefunction([DACEDA_p, c_double])
def CreateFilled(ina, ckon: float) -> None:
    """`void daceCreateFilled(DACEDA REF(ina), const double ckon);`"""
    ...


@corefunction([DACEDA_p, c_double])
def CreateRandom(ina, cm: float) -> None:
    """`void daceCreateRandom(DACEDA REF(ina), const double cm);`"""
    ...


# *****************************************************************************
# *     DACE coefficient access routines
# *****************************************************************************


@corefunction([DACEDA_p], c_double)
def GetConstant(ina) -> float:
    """`double daceGetConstant(const DACEDA REF(ina));`"""
    ...


@corefunction([DACEDA_p, c_double_p])
def GetLinear(ina, c) -> None:
    """`void daceGetLinear(const DACEDA REF(ina), double c[]);`"""
    ...


@corefunction([DACEDA_p, c_uint_p], c_double)
def GetCoefficient(ina, jj) -> float:
    """`double daceGetCoefficient(const DACEDA REF(ina), const unsigned int jj[]);`"""
    ...


@corefunction([DACEDA_p, c_uint], c_double)
def GetCoefficient0(ina, ic: int) -> float:
    """`double daceGetCoefficient0(const DACEDA REF(ina), const unsigned int ic);`"""
    ...


@corefunction([DACEDA_p, c_uint_p, c_double])
def SetCoefficient(ina, jj, cjj: float) -> None:
    """`void daceSetCoefficient(DACEDA REF(ina), const unsigned int jj[], const double cjj);`"""
    ...


@corefunction([DACEDA_p, c_uint, c_double])
def SetCoefficient0(ina, ic: int, cjj: float) -> None:
    """`void daceSetCoefficient0(DACEDA REF(ina), const unsigned int ic, const double cjj);`"""
    ...


@corefunction([DACEDA_p, c_uint, c_uint_p, c_double_p])
def GetCoefficientAt(ina, npos: int, jj, cjj) -> None:
    """`void daceGetCoefficientAt(const DACEDA REF(ina), const unsigned int npos, unsigned int jj[], double REF(cjj));`"""
    ...


@corefunction([DACEDA_p], c_uint)
def GetLength(ina) -> int:
    """`unsigned int daceGetLength(const DACEDA REF(ina));`"""
    ...


# *****************************************************************************
# *     DACE DA copying and filtering
# *****************************************************************************


@corefunction([DACEDA_p, DACEDA_p])
def Copy(ina, inb) -> None:
    """`void daceCopy(const DACEDA REF(ina), DACEDA REF(inb));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def CopyFiltering(ina, inb) -> None:
    """`void daceCopyFiltering(const DACEDA REF(ina), DACEDA REF(inb));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p, DACEDA_p])
def Filter(ina, inb, inc) -> None:
    """`void daceFilter(const DACEDA REF(ina), DACEDA REF(inb), const DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_uint, c_uint, DACEDA_p])
def Trim(ina, imin: int, imax: int, inc) -> None:
    """`void daceTrim(const DACEDA REF(ina), const unsigned int imin, const unsigned int imax, DACEDA REF(inc));`"""
    ...


# *****************************************************************************
# *     Basic DACE arithmetic operations
# *****************************************************************************


@corefunction([DACEDA_p, c_double, DACEDA_p, c_double, DACEDA_p])
def WeightedSum(ina, afac: float, inb, bfac: float, inc) -> None:
    """`void daceWeightedSum(const DACEDA REF(ina), const double afac, const DACEDA REF(inb), const double bfac, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p, DACEDA_p])
def Add(ina, inb, inc) -> None:
    """`void daceAdd(const DACEDA REF(ina), const DACEDA REF(inb), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p, DACEDA_p])
def Subtract(ina, inb, inc) -> None:
    """`void daceSubtract(const DACEDA REF(ina), const DACEDA REF(inb), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p, DACEDA_p])
def Multiply(ina, inb, inc) -> None:
    """`void daceMultiply(const DACEDA REF(ina), const DACEDA REF(inb), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p, DACEDA_p])
def MultiplyMonomials(ina, inb, inc) -> None:
    """`void daceMultiplyMonomials(const DACEDA REF(ina), const DACEDA REF(inb), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p, DACEDA_p])
def Divide(ina, inb, inc) -> None:
    """`void daceDivide(const DACEDA REF(ina), const DACEDA REF(inb), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def Square(ina, inb) -> None:
    """`void daceSquare(const DACEDA REF(ina), DACEDA REF(inb));`"""
    ...


@corefunction([DACEDA_p, c_double, DACEDA_p])
def AddDouble(ina, ckon: float, inb) -> None:
    """`void daceAddDouble(const DACEDA REF(ina), const double ckon, DACEDA REF(inb));`"""
    ...


@corefunction([DACEDA_p, c_double, DACEDA_p])
def DoubleSubtract(ina, ckon: float, inb) -> None:
    """`void daceDoubleSubtract(const DACEDA REF(ina), const double ckon, DACEDA REF(inb));`"""
    ...


@corefunction([DACEDA_p, c_double, DACEDA_p])
def SubtractDouble(ina, ckon: float, inb) -> None:
    """`void daceSubtractDouble(const DACEDA REF(ina), const double ckon, DACEDA REF(inb));`"""
    ...


@corefunction([DACEDA_p, c_double, DACEDA_p])
def MultiplyDouble(ina, ckon: float, inb) -> None:
    """`void daceMultiplyDouble(const DACEDA REF(ina), const double ckon, DACEDA REF(inb));`"""
    ...


@corefunction([DACEDA_p, c_double, DACEDA_p])
def DivideDouble(ina, ckon: float, inb) -> None:
    """`void daceDivideDouble(const DACEDA REF(ina), const double ckon, DACEDA REF(inb));`"""
    ...


@corefunction([DACEDA_p, c_double, DACEDA_p])
def DoubleDivide(ina, ckon: float, inb) -> None:
    """`void daceDoubleDivide(const DACEDA REF(ina), const double ckon, DACEDA REF(inb));`"""
    ...


@corefunction([DACEDA_p, c_uint, c_uint, DACEDA_p])
def DivideByVariable(ina, var: int, p: int, inc) -> None:
    """`void daceDivideByVariable(const DACEDA REF(ina), const unsigned int var, const unsigned int p, DACEDA REF(inc));`"""
    ...


@corefunction([c_uint, DACEDA_p, DACEDA_p])
def Differentiate(idif: int, ina, inc) -> None:
    """`void daceDifferentiate(const unsigned int idif, const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([c_uint, DACEDA_p, DACEDA_p])
def Integrate(iint: int, ina, inc) -> None:
    """`void daceIntegrate(const unsigned int iint, const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


# *****************************************************************************
# *     DACE intrinsic function routines
# *****************************************************************************


@corefunction([DACEDA_p, DACEDA_p])
def Truncate(ina, inc) -> None:
    """`void daceTruncate(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def Round(ina, inc) -> None:
    """`void daceRound(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_double, DACEDA_p])
def Modulo(ina, p: float, inc) -> None:
    """`void daceModulo(const DACEDA REF(ina), const double p, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_double, DACEDA_p])
def PowerDouble(ina, p: float, inc) -> None:
    """`void dacePowerDouble(const DACEDA REF(ina), const double p, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_int, DACEDA_p])
def Power(ina, np: int, inc) -> None:
    """`void dacePower(const DACEDA REF(ina), const int np, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_int, DACEDA_p])
def Root(ina, np: int, inc) -> None:
    """`void daceRoot(const DACEDA REF(ina), const int np, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def MultiplicativeInverse(*args) -> None:
    """`void daceMultiplicativeInverse(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def SquareRoot(ina, inc) -> None:
    """`void daceSquareRoot(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def InverseSquareRoot(ina, inc) -> None:
    """`void daceInverseSquareRoot(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def CubicRoot(ina, inc) -> None:
    """`void daceCubicRoot(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def InverseCubicRoot(ina, inc) -> None:
    """`void daceInverseCubicRoot(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p, DACEDA_p])
def Hypotenuse(ina, inb, inc) -> None:
    """`void daceHypotenuse(const DACEDA REF(ina), const DACEDA REF(inb), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def Exponential(ina, inc) -> None:
    """`void daceExponential(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def Logarithm(ina, inc) -> None:
    """`void daceLogarithm(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_double, DACEDA_p])
def LogarithmBase(ina, b: float, inc) -> None:
    """`void daceLogarithmBase(const DACEDA REF(ina), const double b, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def Logarithm10(ina, inc) -> None:
    """`void daceLogarithm10(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def Logarithm2(ina, inc) -> None:
    """`void daceLogarithm2(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def Sine(ina, inc) -> None:
    """`void daceSine(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def Cosine(ina, inc) -> None:
    """`void daceCosine(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def Tangent(ina, inc) -> None:
    """`void daceTangent(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def ArcSine(ina, inc) -> None:
    """`void daceArcSine(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def ArcCosine(ina, inc) -> None:
    """`void daceArcCosine(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def ArcTangent(ina, inc) -> None:
    """`void daceArcTangent(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p, DACEDA_p])
def ArcTangent2(*args) -> None:
    """`void daceArcTangent2(const DACEDA REF(ina), const DACEDA REF(inb), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def HyperbolicSine(ina, inc) -> None:
    """`void daceHyperbolicSine(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def HyperbolicCosine(ina, inc) -> None:
    """`void daceHyperbolicCosine(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def HyperbolicTangent(ina, inc) -> None:
    """`void daceHyperbolicTangent(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def HyperbolicArcSine(ina, inc) -> None:
    """`void daceHyperbolicArcSine(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def HyperbolicArcCosine(ina, inc) -> None:
    """`void daceHyperbolicArcCosine(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def HyperbolicArcTangent(ina, inc) -> None:
    """`void daceHyperbolicArcTangent(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def ErrorFunction(ina, inc) -> None:
    """`void daceErrorFunction(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def ComplementaryErrorFunction(ina, inc) -> None:
    """`void daceComplementaryErrorFunction(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_int, c_bool, DACEDA_p])
def BesselIFunction(ina, n: int, scaled: bool, inc) -> None:
    """`void daceBesselIFunction(const DACEDA REF(ina), const int n, const bool scaled, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_int, DACEDA_p])
def BesselJFunction(ina, n: int, inc) -> None:
    """`void daceBesselJFunction(const DACEDA REF(ina), const int n, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_int, c_bool, DACEDA_p])
def BesselKFunction(ina, n: int, scaled: bool, inc) -> None:
    """`void daceBesselKFunction(const DACEDA REF(ina), const int n, const bool scaled, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_int, DACEDA_p])
def BesselYFunction(ina, n: int, inc) -> None:
    """`void daceBesselYFunction(const DACEDA REF(ina), const int n, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def LogGammaFunction(ina, inc) -> None:
    """`void daceLogGammaFunction(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, DACEDA_p])
def GammaFunction(ina, inc) -> None:
    """`void daceGammaFunction(const DACEDA REF(ina), DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_uint, DACEDA_p])
def PsiFunction(ina, n: int, inc) -> None:
    """`void dacePsiFunction(const DACEDA REF(ina), const unsigned int n, DACEDA REF(inc));`"""
    ...


# *****************************************************************************
# *     DACE norm and norm estimation routines
# *****************************************************************************


@corefunction([DACEDA_p], c_double)
def AbsoluteValue(ina) -> float:
    """`double daceAbsoluteValue(const DACEDA REF(ina));`"""
    ...


@corefunction([DACEDA_p, c_uint], c_double)
def Norm(ina, ityp: int) -> float:
    """`double daceNorm(const DACEDA REF(ina), const unsigned int ityp);`"""
    ...


@corefunction([DACEDA_p, c_uint, c_uint, c_double_p])
def OrderedNorm(ina, ivar: int, ityp: int, onorm) -> None:
    """`void daceOrderedNorm(const DACEDA REF(ina), const unsigned int ivar, const unsigned int ityp, double onorm[]);`"""
    ...


@corefunction([DACEDA_p, c_uint, c_uint, c_double_p, c_double_p, c_uint])
def Estimate(ina, ivar: int, ityp: int, c, err, nc: int) -> None:
    """`void daceEstimate(const DACEDA REF(ina), const unsigned int ivar, const unsigned int ityp, double c[], double err[], const unsigned int nc);`"""
    ...


@corefunction([DACEDA_p, c_double_p, c_double_p])
def GetBounds(ina, alo, aup) -> None:
    """`void daceGetBounds(const DACEDA REF(ina), double REF(alo), double REF(aup));`"""
    ...


# *****************************************************************************
# *     DACE polynomial evaluation routines
# *****************************************************************************


@corefunction([DACEDA_p, DACEDA_p], c_double)
def EvalMonomials(ina, inb) -> float:
    """`double daceEvalMonomials(const DACEDA REF(ina), const DACEDA REF(inb));`"""
    ...


@corefunction([DACEDA_p, c_uint, c_uint, c_double, DACEDA_p])
def ReplaceVariable(ina, from_: int, to: int, val: float, inc) -> None:
    """`void daceReplaceVariable(const DACEDA REF(ina), const unsigned int from, const unsigned int to, const double val, DACEDA REF(inc));`""",
    ...


@corefunction([DACEDA_p, c_uint, c_double, DACEDA_p])
def EvalVariable(ina, nvar: int, val: float, inc) -> None:
    """`void daceEvalVariable(const DACEDA REF(ina), const unsigned int nvar, const double val, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_uint, c_double, DACEDA_p])
def ScaleVariable(ina, nvar: int, val: float, inc) -> None:
    """`void daceScaleVariable(const DACEDA REF(ina), const unsigned int nvar, const double val, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p, c_uint, c_double, c_double, DACEDA_p])
def TranslateVariable(ina, nvar: int, a: float, c: float, inc) -> None:
    """`void daceTranslateVariable(const DACEDA REF(ina), const unsigned int nvar, const double a, const double c, DACEDA REF(inc));`"""
    ...


@corefunction([DACEDA_p_p, c_uint, c_double_p, c_uint_p, c_uint_p, c_uint_p])
def EvalTree(das, count: int, ac, nterm, nvar, nord) -> None:
    """`void daceEvalTree(const DACEDA *das[], const unsigned int count, double ac[], unsigned int REF(nterm), unsigned int REF(nvar), unsigned int REF(nord));`"""
    ...


# *****************************************************************************
# *     DACE input/output routines
# *****************************************************************************


@corefunction([DACEDA_p, c_char_p, c_uint_p])
def Write(ina, strs, nstrs) -> None:
    """`void daceWrite(const DACEDA REF(ina), char *strs, unsigned int REF(nstrs));`"""
    ...


@corefunction([DACEDA_p, c_char_p, c_uint])
def Read(ina, strs, nstrs) -> None:
    """`void daceRead(DACEDA REF(ina), char *strs, unsigned int nstrs);`"""
    ...


@corefunction([DACEDA_p])
def Print(ina) -> None:
    """`void dacePrint(const DACEDA REF(ina));`"""
    ...


@corefunction([DACEDA_p, c_void_p, c_uint_p], c_uint)
def ExportBlob(ina, blob, size) -> int:
    """`unsigned int daceExportBlob(const DACEDA REF(ina), void *blob, unsigned int REF(size));`"""
    ...


@corefunction([c_void_p], c_uint)
def BlobSize(blob) -> int:
    """`unsigned int daceBlobSize(const void *blob);`"""
    ...


@corefunction([c_void_p, DACEDA_p])
def ImportBlob(blob, inc) -> None:
    """`void daceImportBlob(const void *blob, DACEDA REF(inc));`"""
    ...


# *****************************************************************************
# *     DACE miscellaneous routines
# *****************************************************************************


@corefunction([], c_double)
def Random() -> float:
    """`double daceRandom();`"""
    ...



# From: https://stackoverflow.com/a/71551501

import platform
import sysconfig
import sys


def get_platform():
    """
    Return a string with current platform (system and machine architecture).

    This attempts to improve upon `sysconfig.get_platform` by fixing some
    issues when running a Python interpreter with a different architecture than
    that of the system (e.g. 32bit on 64bit system, or a multiarch build),
    which should return the machine architecture of the currently running
    interpreter rather than that of the system (which didn't seem to work
    properly). The reported machine architectures follow platform-specific
    naming conventions (e.g. "x86_64" on Linux, but "x64" on Windows).

    Example output strings for common platforms:

        darwin_(ppc|ppc64|i368|x86_64|arm64)
        linux_(i686|x86_64|armv7l|aarch64)
        windows_(x86|x64|arm32|arm64)

    """
    system = platform.system().lower()
    machine = sysconfig.get_platform().split("-")[-1].lower()
    is_64bit = sys.maxsize > 2 ** 32
    if system == "darwin":  # get machine architecture of multiarch binaries
        if any(x in machine for x in ("fat", "intel", "universal")):
            machine = platform.machine().lower()
    elif system == "linux":  # fix running 32bit interpreter on 64bit system
        if not is_64bit and machine == "x86_64":
            machine = "i686"
        elif not is_64bit and machine == "aarch64":
            machine = "armv7l"
    elif system == "windows":  # return more precise machine architecture names
        if machine == "amd64":
            machine = "x64"
        elif machine == "win32":
            if is_64bit:
                machine = platform.machine().lower()
            else:
                machine = "x86"
    # some more fixes based on examples in https://en.wikipedia.org/wiki/Uname
    if not is_64bit and machine in ("x86_64", "amd64"):
        if any(x in system for x in ("cygwin", "mingw", "msys")):
            machine = "i686"
        else:
            machine = "i386"
    return system, machine



"""
# Differential Algebra Core Engine in Python - DACEyPy

DACEyPy is a Python wrapper of DACE,
the Differential Algebra Computational Toolbox
(https://github.com/dacelib/dace).

------------------------------------------------------------------------------

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

------------------------------------------------------------------------------
"""

__all__ = (
    "__version__",
    "DACEException",
    "DA",
    "compiledDA",
    "array",
    "Monomial",
    "init",
    "isInitialized",
    "zeros",
    "identity",
    "op",
    "ADS",
    "RK",
    "integrator",
    "ADSintegrator",
)

from ._version import __version__
from ._DACEException import DACEException
from ._DA import DA
from ._compiledDA import compiledDA
from ._array import array
from ._Monomial import Monomial
from . import op
from ._ADS import ADS
from . import RK
from ._integrator import integrator
from ._ADSintegrator import ADSintegrator

init = DA.init
isInitialized = DA.isInitialized
zeros = array.zeros
identity = array.identity
